using System;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties

		private readonly Face[] faces;
		/// <summary>Gets the faces for the level.</summary>
		[Description("All of the faces in the level.")]
		[Category("Geometry")]
		public Face[] Faces {
			get { return this.faces; }
		}

		#endregion

		/// <summary>Represents a convex polygon, made up of edges.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Face {

			#region Types

			/// <summary>Specifies the side of a plane the face is on.</summary>
			public enum PlaneSide : ushort {
				/// <summary>The visible side of the face lies in front of the plane.</summary>
				Front = 0,
				/// <summary>The visible side of the face lies behind the plane.</summary>
				Back = 1,
			}

			/// <summary>Represents a light map which can be applied to a face.</summary>
			public class LightMap : IPicture {

				#region Properties

				private readonly Face face;
				/// <summary>The face the light map is applied to.</summary>
				public Face Face {
					get { return this.face; }
				}

				private readonly int width;
				/// <summary>The width of the lightmap.</summary>
				public int Width {
					get { return this.width; }
				}

				private readonly int height;
				/// <summary>The height of the lightmap.</summary>
				public int Height {
					get { return this.height; }
				}

				private readonly byte[] data;
				/// <summary>The lightmap data.</summary>
				public byte[] Data {
					get { return this.data; }
				}


				#endregion

				#region Constructor

				public LightMap(Face face, int width, int height, byte[] data) {
					this.face = face;
					this.width = width; this.height = height;
					this.data = data;
				}

				#endregion

				#region Methods

				/// <summary>Gets a normalised (0..1) texture coordinate for a vertex on the face.</summary>
				/// <param name="position">The position of the vertex to get the texture coordinates for.</param>
				/// <returns>A normalised texture coordinate.</returns>
				public Vector2 GetNormalisedTextureCoordinate(Vector3 position) {
					return ((this.Face.Get2DFrom3D(position)) - this.Face.TextureMin) / (this.Face.Extent + new Vector2(16));
				}

				#endregion

				#region IPicture Members

				public Picture Picture {
					get {
						int[] ExpandedData = new int[this.Width * this.Height];
						switch (this.Face.Level.FileVersion.LightMapBytesPerPixel) {
							case 1:
								ExpandedData = Array.ConvertAll<byte, int>(this.data, delegate(byte b) { return unchecked((int)0xFF000000 | b | (b << 8) | (b << 16)); });
								break;
							case 3:
								for (int i = 0, j = 0; i < ExpandedData.Length; i++) {
									ExpandedData[i] = unchecked((int)0xFF000000 | (this.data[j++] << 16) | (this.data[j++] << 8) | (this.data[j++] << 0));
								}
								break;
							default:
								throw new InvalidOperationException("Unsupported light map version.");
						}
						return new Picture(this.width, this.height, ExpandedData);
					}
				}

				#endregion
			}

			#endregion

			#region Methods

			/// <summary>Get the vertex indices for the face correctly ordered as a convex polygon.</summary>
			public ushort[] GetVertexIndices() {
				ushort[] Indices = new ushort[this.EdgeCount];
				uint EdgeIndex = this.EdgeListIndex;
				for (int i = 0; i < this.EdgeCount; ++i) {
					int ActualEdgeIndex = this.Level.EdgeList[EdgeIndex++];
					Edge Edge = this.Level.Edges[Math.Abs(ActualEdgeIndex)];
					if (ActualEdgeIndex > 0) {
						Indices[i] = Edge.Start;
					} else {
						Indices[i] = Edge.End;
					}
				}
				return Indices;
			}


			/// <summary>Gets a 2D position on the face by removing the constant component of a 3D position.</summary>
			/// <param name="position">The 3D position to convert.</param>
			/// <returns>The 2D position.</returns>
			public Vector2 Get2DFrom3D(Vector3 position) {
				TextureInfo T = this.level.TextureInformation[this.Texture];
				return new 
					Vector2(
						Vector3.Dot(position, level.TextureInformation[this.Texture].Horizontal),
						Vector3.Dot(position, level.TextureInformation[this.Texture].Vertical)
					) + new Vector2(T.HorizontalOffset, T.VerticalOffset);
			}


			/// <summary>Calculate the intersection point of a point moving through the face.</summary>
			/// <param name="start">The starting position of the point.</param>
			/// <param name="end">The desired end position of the point.</param>
			/// <returns>Either the end point or the intersection point, depending on whether the point hits the face or not.</returns>
			public CollisionResult<Vector3> GetCollision(Vector3 start, Vector3 end) {

				CollisionResult<Vector3> C = new CollisionResult<Vector3>(false, end);

				// Start by taking a bounding box around the ray.
				BoundingBox RayBounds = new BoundingBox(Vector3.Min(start, end), Vector3.Max(start, end));

				// Check to see if the bounding box intersects with our one.
				if (!RayBounds.Intersects(this.bounds)) {
					// The ray completely misses us (hurrah!)
					return C;
				}

				// Now, it's not so easy. Clip against the plane, for starters:

				Level.Plane SourcePlane = this.level.planes[this.plane];
				CollisionResult<Vector3>PlaneCollision = SourcePlane.GetCollision(start, end);


				// If we reach the end of line anyway, all is good.
				if (!PlaneCollision.CollisionOccured) return PlaneCollision;

				ushort[] VertexIndices = this.GetVertexIndices();

				Vector3 ActualNormal = this.side == PlaneSide.Front ? SourcePlane.Normal : -SourcePlane.Normal;
				bool LastDirection = false;

				// At this stage we need to test whether the point of intersection hits the face inside the bounding box,
				// or whether it misses it.
				for (int i = 0; i < VertexIndices.Length; ++i) {
					
					Vector3 Edge0 = this.level.vertices[VertexIndices[i]];
					Vector3 Edge1 = this.level.vertices[VertexIndices[(i + 1) % VertexIndices.Length]];

					bool AlongNormal = Vector3.Dot(this.Normal, Vector3.Cross(PlaneCollision.Position - Edge0, Edge1 - Edge0)) > 0;
					if (i == 0) LastDirection = AlongNormal;

					if (LastDirection != AlongNormal) return C;

					LastDirection = AlongNormal;
				}


				// We have hit the face!
				PlaneCollision.Face = this;
				return PlaneCollision;
			}

			/// <summary>Get the collision position between a bounding box and a face.</summary>
			/// <param name="box">The bounding box to move.</param>
			/// <param name="direction">The direction vector to move the box by.</param>
			public CollisionResult<BoundingBox> GetCollision(BoundingBox box, Vector3 direction) {
				
				BoundingBox End = box + direction;
				BoundingBox TotalBoxBounds = BoundingBox.Combine(End, box);
				if (!this.bounds.Intersects(TotalBoxBounds) ||                                    // Quick bounding-box test.
					!((Quake.Plane)(this.level.planes[this.plane])).Intersects(TotalBoxBounds)) { // Plane intersection test.
					return new CollisionResult<BoundingBox>(false, End);
				}

				// Urp. The bounding boxes and plane collide.
				
				Vector3[] Starts = box.Corners;
				Vector3[] Ends = End.Corners;

				//OPTIMISATION: Only check corners on sides that are moving in direction of bounding box.

				// Perform collision detection for all eight corners:
				CollisionResult<BoundingBox> Result = new CollisionResult<BoundingBox>(false, End);
				for (int i = 0; i < 8; ++i) {
					CollisionResult<Vector3> C = this.GetCollision(Starts[i], Ends[i]); // Accurate point->face collision.
					if (C.CollisionOccured && C.Distance < Result.Distance) {
						Result.CollisionOccured = true;
						Result.Distance = C.Distance;
					}
				}

				// If a collision occured...
				if (Result.CollisionOccured) {
					Result.Face = this;
					Result.Plane = (Quake.Plane)(this.level.planes[this.plane]);
					Result.Position = box + direction * Result.Distance;
				}

				return Result;
			}


			private LightMap lightMap;

			/// <summary>Gets a light map for the face.</summary>
			/// <returns>A light map for the face.</returns>
			public LightMap GetLightMap() {
				if (this.lightMap == null) {
					if (!this.HasLightMap) {

						byte Brightness = (byte)(this.Level.TextureInformation[this.Texture].IsLiquid ? 0x80 : (0xFF - this.Darkness));
						byte[] Data = new byte[this.Level.FileVersion.LightMapBytesPerPixel];
						for (int i = 0; i < Data.Length; ++i) Data[i] = Brightness;
						this.lightMap = new LightMap(this, 1, 1, Data);

					} else {

						int Width = ((int)this.Extent.X >> 4) + 1;
						int Height = ((int)this.Extent.Y >> 4) + 1;

						byte[] Data = new byte[Width * Height * this.Level.FileVersion.LightMapBytesPerPixel];
						level.BaseStream.Seek(level.Entries[level.FileVersion.GetEntryIndex(Entry.EntryTypes.LightMaps)].Offset + this.LightMapData, SeekOrigin.Begin);
						level.BaseStream.Read(Data, 0, Data.Length);

						this.lightMap = new LightMap(this, Width, Height, Data);
					}
				}
				return this.lightMap;
			}

			#endregion

			#region Constructor

			public Face(Level level, Stream stream) {
				this.level = level;

				BinaryReader WorkingReader = new BinaryReader(stream);

				this.plane = WorkingReader.ReadUInt16();
				this.side = (Face.PlaneSide)WorkingReader.ReadUInt16();
				this.edgeListIndex = WorkingReader.ReadUInt32();
				this.edgeCount = WorkingReader.ReadUInt16();
				this.texture = WorkingReader.ReadUInt16();

				this.lightStyle = WorkingReader.ReadByte();
				this.darkness = WorkingReader.ReadByte();
				this.lightData = WorkingReader.ReadBytes(2);

				this.lightMapData = WorkingReader.ReadUInt32();

				// Calculate extents:

				Vector2 Min2D = new Vector2(float.MaxValue, float.MaxValue);
				Vector2 Max2D = new Vector2(float.MinValue, float.MinValue);

				for (int i = 0; i < this.EdgeCount; ++i) {
					int EdgeListIndex = level.EdgeList[this.EdgeListIndex + i];
					Vector3 v;
					if (EdgeListIndex >= 0) {
						v = level.vertices[level.edges[+EdgeListIndex].Start];
					} else {
						v = level.vertices[level.edges[-EdgeListIndex].End];
					}
					Vector2 TestCoordinate = Get2DFrom3D(v);
					Min2D = Vector2.Min(TestCoordinate, Min2D);
					Max2D = Vector2.Max(TestCoordinate, Max2D);
				}


				Min2D.X = (float)Math.Floor(Min2D.X / 16f);
				Min2D.Y = (float)Math.Floor(Min2D.Y / 16f);
				Max2D.X = (float)Math.Ceiling(Max2D.X / 16f);
				Max2D.Y = (float)Math.Ceiling(Max2D.Y / 16f);

				this.textureMin = (Min2D - 0.5f) * 16f;
				this.extent = (Max2D - Min2D) * 16f;

				long l = stream.Position;
				this.GetLightMap();
				stream.Position = l;

				// Get bounding box:
				Vector3 Min = new Vector3(float.MaxValue);
				Vector3 Max = new Vector3(float.MinValue);
				foreach (ushort vi in this.GetVertexIndices()) {
					Vector3 Vertex = this.level.vertices[vi];
					Min = Vector3.Min(Min, Vertex);
					Max = Vector3.Max(Max, Vertex);
				}
				this.bounds = new BoundingBox(Min, Max);

			}

			#endregion			

			#region Properties

			private Level level;
			[Browsable(false)]
			public Level Level {
				get { return this.level; }
			}


			private readonly ushort plane;
			/// <summary>Index of the plane in which the face lies.</summary>
			[Description("Index of the plane in which the face lies.")]
			[Category("Geometry")]
			public ushort Plane {
				get { return this.plane; }
			}

			private PlaneSide side;
			/// <summary>The side of a plane the face is on.</summary>
			[Description("The side of a plane the face is on.")]
			[Category("Geometry")]
			public PlaneSide Side {
				get { return this.side; }
			}

			private uint edgeListIndex;
			/// <summary>The index of the first edge of the face.</summary>
			[Description("The index of the first edge of the face.")]
			[Category("Geometry")]
			public uint EdgeListIndex {
				get { return this.edgeListIndex; }
			}

			private ushort edgeCount;
			/// <summary>How many edges make up the face.</summary>
			[Description("How many edges make up the face.")]
			[Category("Geometry")]
			public ushort EdgeCount {
				get { return this.edgeCount; }
			}

			private ushort texture;
			/// <summary>Texture index for the face.</summary>
			[Description("Texture index for the face.")]
			[Category("Texture")]
			public ushort Texture {
				get { return this.texture; }
			}

			private byte lightStyle;
			/// <summary>The type of lighting for the face.</summary>
			[Description("The type of lighting for the face.")]
			[Category("Lighting")]
			public byte LightStyle {
				get { return this.lightStyle; }
			}

			private byte darkness;
			/// <summary>Base light value, from 0x00 (light) to 0xFF (dark).</summary>
			[Description("Base light value, from 0x00 (light) to 0xFF (dark).")]
			[Category("Lighting")]
			public byte Darkness {
				get { return this.darkness; }
			}

			private byte[] lightData;
			/// <summary>Additional light data.</summary>
			[Description("Additional light data.")]
			[Category("Lighting")]
			public byte[] LightData {
				get { return this.lightData; }
			}

			private uint lightMapData;
			/// <summary>Pointer to light map info or uint.MaxValue if no light map is available.</summary>
			[Description("Pointer to light map info or uint.MaxValue if no light map is available.")]
			[Category("Lighting")]
			public uint LightMapData {
				get { return this.lightMapData; }
			}

			/// <summary>Gets whether the face has a light map associated with it.</summary>
			[Description("Whether the face has a light map associated with it.")]
			[Category("Lighting")]
			public bool HasLightMap {
				get {
					return 
						this.LightMapData != uint.MaxValue && 
						!this.Level.TextureInformation[this.Texture].IsSky &&
						!this.Level.TextureInformation[this.Texture].IsLiquid;
				}
			}


			private Vector2 textureMin;
			/// <summary>Gets the smallest texture coordinate for the face.</summary>
			[Description("The smallest texture coordinate for the face.")]
			[Category("Texture")]
			public Vector2 TextureMin {
				get { return this.textureMin; }
			}

			private Vector2 extent;
			/// <summary>Gets the extent of the face's texture coordinates rounded up to the nearest 16 pixels.</summary>
			[Description("The extent of the face's texture coordinates rounded up to the nearest 16 pixels.")]
			[Category("Texture")]
			public Vector2 Extent {
				get { return this.extent; }
			}

			private readonly BoundingBox bounds;
			/// <summary>Gets the bounding box  of the entire face.</summary>
			[Description("The bounding box  of the entire face.")]
			[Category("Geometry")]
			public BoundingBox Bounds {
				get { return this.bounds; }
			}


			public override string ToString() {
				return this.EdgeCount.ToString() + " Edge" + (this.EdgeCount == 1 ? "" : "s");
			}

			/// <summary>Gets the normal of this face pointing the correct way, according to the side of the plane it's on.</summary>
			public Vector3 Normal {
				get { return this.Level.planes[this.plane].Normal * (this.side == PlaneSide.Front ? +1 : -1); }
			}


			#endregion
		

		}

	}
}
