using System;
using System.Collections.Generic;
using System.IO;
using System.ComponentModel;

namespace id.Quake.Files {
	public partial class Level {

		#region Properties


		private readonly Leaf[] leaves;
		/// <summary>Gets the level's BSP leaves.</summary>
		[Category("BSP Tree")]
		[Description("All of the leaves in the level.")]
		public Leaf[] Leaves {
			get { return this.leaves; }
		}

		#endregion


		/// <summary>Defines a leaf on the BSP tree.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Leaf {

			#region Types

			/// <summary>Represents different Quake 1 special leaf types.</summary>
			public enum Quake1Types {
				/// <summary>Contains normal "air".</summary>
				Normal = -1,
				/// <summary>Contains a solid region (such as a wall) or is outside the level entirely.</summary>
				InsideSolid = -2,
				/// <summary>Contains water.</summary>
				Water = -3,
				/// <summary>Contains slime.</summary>
				Slime = -4,
				/// <summary>Contains lava.</summary>
				Lava = -5,
				/// <summary>Contains the sky.</summary>
				Sky = -6,
			}

			/// <summary>Represents different Quake 2 special leaf types.</summary>
			[Flags()] public enum Quake2Types {

				None = 0x00000000,
				Solid = 0x00000001,
				Window = 0x00000002,
				Aux = 0x00000004,
				Lava = 0x00000008,
				Slime = 0x00000010,
				Water = 0x00000020,
				Mist = 0x00000040,

				AreaPortal = 0x00008000,

				PlayerClip = 0x00010000,
				MonsterClip = 0x00020000,

				Current0 = 0x00040000,
				Current90 = 0x00080000,
				Current180 = 0x00100000,
				Current270 = 0x00200000,
				CurrentUp = 0x00400000,
				CurrentDown = 0x00800000,

				Origin = 0x01000000,

				Monster = 0x02000000,
				DeadMonster = 0x04000000,
				Detail = 0x08000000,
				Translucent = 0x08000000,
				Ladder = 0x10000000,
			}
			
			#endregion

			#region Public Properties

			private readonly Level level;
			/// <summary>Gets the level that contains this leaf.</summary>
			public Level Level {
				get { return this.level; }
			}

			private readonly uint type;

			/// <summary>Quake 1 specific leaf type.</summary>
			public Quake1Types Quake1Type {
				get { return (Quake1Types)type; }
			}

			/// <summary>Quake 2 specific leaf type.</summary>
			public Quake2Types Quake2Type {
				get { return (Quake2Types)type; }
			}

			private readonly uint visibilityList;
			/// <summary>Pointer to this leaf's visibility list.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public uint VisibilityList { get { this.Level.FileVersion.BreakOnInvalidGameVersion(Version.Game.Quake); return this.visibilityList; } }

			/// <summary>Index of this leaf's visibility cluster.</summary>
			/// <remarks>Quake 2 only.</remarks>
			public ushort VisibilityCluster { get { this.Level.FileVersion.BreakOnInvalidGameVersion(Version.Game.Quake2); return (ushort)(this.visibilityList); } }

			/// <summary>Visibility area.</summary>
			/// <remarks>Quake 2 only.</remarks>
			public ushort VisibilityArea { get { this.Level.FileVersion.BreakOnInvalidGameVersion(Version.Game.Quake2); return (ushort)(this.visibilityList >> 16); } }


			private readonly BoundingBoxShort bounds;
			/// <summary>Packed bounding box of the leaf.</summary>
			public BoundingBoxShort Bounds { get { return this.bounds; } }

			private readonly ushort faceListIndex;
			/// <summary>Index of the first face in this leaf.</summary>
			public ushort FaceListIndex { get { return this.faceListIndex; } }

			private readonly ushort faceCount;
			/// <summary>Number of faces in this leaf.</summary>
			public ushort FaceCount { get { return this.faceCount; } }

			private readonly byte[] effectVolumes;

			/// <summary>Volume of the water ambient sound.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public byte VolumeWater { get { return this.effectVolumes[0]; } }
			/// <summary>Volume of the sky ambient sound.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public byte VolumeSky { get { return this.effectVolumes[1]; } }
			/// <summary>Volume of the slime ambient sound.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public byte VolumeSlime { get { return this.effectVolumes[2]; } }
			/// <summary>Volume of the lava ambient sound.</summary>
			/// <remarks>Quake 1 only.</remarks>
			public byte VolumeLava { get { return this.effectVolumes[3]; } }

			#endregion

			#region Extended Properties

			/// <summary>Returns true if the leaf has a visibility list.</summary>
			public bool HasVisibilityList {
				get { return this.Level.FileVersion.GameVersion == Version.Game.Quake ? this.VisibilityList != uint.MaxValue : this.VisibilityCluster != ushort.MaxValue; }
			}

			/// <summary>Gets the type as a generic Quake 2 type.</summary>
			/// <remarks>Quake 1 types are converted to Quake 2 types.</remarks>
			public Quake2Types Type {
				get {
					switch (this.Level.FileVersion.GameVersion) {
						case Version.Game.Quake:
							switch (this.Quake1Type) {
								case Quake1Types.InsideSolid: return Quake2Types.Solid;
								case Quake1Types.Lava: return Quake2Types.Lava;
								case Quake1Types.Normal: return Quake2Types.None;
								case Quake1Types.Sky: return Quake2Types.Window;
								case Quake1Types.Slime: return Quake2Types.Slime;
								case Quake1Types.Water: return Quake2Types.Water;
								default: return Quake2Types.None;
							}
						case Version.Game.Quake2: 
							return this.Quake2Type;
						default: 
							throw new InvalidOperationException();
					}
				}
			}

			/// <summary>Returns a formatted string describing the leaf.</summary>
			public override string ToString() {
				return this.Type.ToString();
			}

			#endregion

			#region Methods

			/// <summary>Get the indices of leaves visible when the camera is in this leaf.</summary>
			/// <param name="model">The model that this leaf is part of.</param>
			/// <returns>Array of indices of visible leaves.</returns>
			public int[] GetVisibilityList(Model model) {

				// Check for visibility lists:
				if (!this.HasVisibilityList) {
					throw new InvalidOperationException("This leaf has no visibility lists.");
				}

				List<int> VisibleLeaves = new List<int>((int)(model.LeafCount / 8));


				int DecompressedQuantity = this.Level.FileVersion.HasVisibilityClusters ? this.Level.VisibilityClusters.Length : model.LeafCount;

				byte[] DecompressedData = GetDecompressedVisibilityList(model, DecompressedQuantity);

				for (int i = 0; i < DecompressedQuantity; ++i) {
					if ((DecompressedData[i >> 3] & (1 << (i & 7))) != 0) {
						if (this.Level.FileVersion.HasVisibilityClusters) {
							VisibleLeaves.AddRange(this.Level.VisibilityClusters[i].VisibleLeaves);
						} else {
							VisibleLeaves.Add(i + 1);
						}
					}
				}


				return VisibleLeaves.ToArray();

			}

			/// <summary>Decompress the visibility list at a particular offset for the model.</summary>
			/// <param name="model">The model to get the visibility list for.</param>
			/// <param name="decompressedQuantity">The quantity of decompressed values.</param>
			/// <remarks>The <c>decompressedQuantity</c> argument refers to the number of leaves in Quake and the number of visibility clusters in Quake 2.</remarks>
			/// <returns>The decompressed data.</returns>
			private byte[] GetDecompressedVisibilityList(Model model, int decompressedQuantity) {
				List<byte> Decompressed = new List<byte>();

				uint Offset = this.Level.FileVersion.HasVisibilityClusters ? this.Level.VisibilityClusters[this.VisibilityCluster].Offset : this.VisibilityList;
				
				if (level.CompressedVisibilityLists.Length > 0) {

					if (Offset != uint.MaxValue) {
						int row = (decompressedQuantity + 7) >> 3;
						do {
							if (level.CompressedVisibilityLists[Offset] != 0) {
								Decompressed.Add(level.CompressedVisibilityLists[Offset++]);
								continue;
							}

							int c = level.CompressedVisibilityLists[Offset + 1];
							Offset += 2;
							while (c != 0) {
								Decompressed.Add(0);
								c--;
							}
						} while (Decompressed.Count < row);
					}
				}
				return Decompressed.ToArray();
			}

			/// <summary>Calculates the intersection point between a start position, end position, and the world.</summary>
			/// <param name="start">The position to start from.</param>
			/// <param name="end">The position to try to end at.</param>
			/// <returns>Information about the collision.</returns>
			/// <remarks>This checks every face in the leaf for a collision.</remarks>
			public CollisionResult<Vector3> GetCollision(Vector3 start, Vector3 end) {
				CollisionResult<Vector3> Result = new CollisionResult<Vector3>(false, end);
				for (int i = 0, j = this.faceListIndex; i < this.faceCount; ++i, ++j) {
					CollisionResult<Vector3> C = this.level.faces[this.level.faceList[j]].GetCollision(start, Result.Position);
					if (C.CollisionOccured) Result = C;
				}
				return Result;
			}

			#endregion

			#region Constructor

			/// <summary>Creates an instance of a leaf.</summary>
			/// <param name="level">The level that the leaf is part of.</param>
			/// <param name="stream">The stream to load the leaf data from.</param>
			public Leaf(Level level, Stream stream) {
				this.level = level;
				BinaryReader WorkingReader = new BinaryReader(stream);
				this.type = WorkingReader.ReadUInt32();
				this.visibilityList = WorkingReader.ReadUInt32();
				this.bounds = new BoundingBoxShort(new Vector3Short(WorkingReader.ReadInt16(), WorkingReader.ReadInt16(), WorkingReader.ReadInt16()), new Vector3Short(WorkingReader.ReadInt16(), WorkingReader.ReadInt16(), WorkingReader.ReadInt16()));
				this.faceListIndex = WorkingReader.ReadUInt16();
				this.faceCount = WorkingReader.ReadUInt16();
				this.effectVolumes = WorkingReader.ReadBytes(4);
			}

			#endregion


		}

	}
}
