using System;
using System.IO;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace id.Quake.Files {
	/// <summary>Represents a compiled QuakeC application.</summary>
	[SupportedExtension("dat")]
	public partial class QuakeC : INamedResource {

		#region Public Properties

		private readonly uint crc;
		/// <summary>Gets the CRC over progdefs.h</summary>
		[Description("Gets the CRC over progdefs.h")]
		public uint Crc { get { return this.crc; } }

		private readonly Statement[] statements;
		/// <summary>Gets all of the statements that make up the application.</summary>
		[Category("Code")]
		[Description("Gets the statements that make up an application.")]
		public Statement[] Statements {
			get { return this.statements; }
		}

		private readonly DefinitionCollection globals;
		/// <summary>Gets the application's global definitions.</summary>
		[Category("Data")]
		[Description("Gets the application's global definitions.")]
		public DefinitionCollection Globals {
			get { return this.globals; }
		}

		private readonly DefinitionCollection fields;
		/// <summary>Gets the application's field definitions.</summary>
		[Category("Data")]
		[Description("Gets the application's field definitions.")]
		public DefinitionCollection Fields {
			get { return this.fields; }
		}

		private readonly FunctionCollection functions;
		/// <summary>Gets the application's functions.</summary>
		[Category("Code")]
		[Description("Gets the application's functions.")]
		public FunctionCollection Functions {
			get { return this.functions; }
		}

		internal string StringBuffer;

		/// <summary>Gets the global variable data values.</summary>
		internal VariableValue[] GlobalData;

		private string name;
		/// <summary>Gets the name of the object.</summary>
		[Browsable(false)] public string Name { get { return this.name; } set { this.name = value; } }

		/// <summary>Gets true if the object contains its own name, false if it uses its filename.</summary>
		[Browsable(false)] public bool NamedInternally { get { return false; } }

		private readonly int edictSize;
		/// <summary>Gets the size of an edict's data in bytes divided by four.</summary>
		public int EdictSize {
			get { return this.edictSize; }
		}

		#endregion

		#region Private Data

		/// <summary>String table within progs.dat</summary>
		private readonly byte[] GlobalStringData;
		/// <summary>Strings as used by the edicts.</summary>
		internal byte[] EdictStringData;

		private readonly ResourceLoader Resources;

		#endregion

		#region Public Methods

		/// <summary>Gets a string from the string table.</summary>
		/// <param name="offset">Offset to the string within the string table.</param>
		internal string GetString(int offset) {

			// If it's uint.MaxValue, that means "use whatever's in the temporary string buffer".
			if (offset == -1) return this.StringBuffer;
			
			// Default to using the string table in PROGS.DAT:
			byte[] StringDataSource = this.GlobalStringData;

			// If the MSB is set, that means "use the string table from the edicts".
			if ((offset & 0x80000000) != 0) {
				StringDataSource = this.EdictStringData;
				offset &= 0x7FFFFFFF;
			}

			// Decode and return the string.
			if (offset > StringDataSource.Length) throw new ArgumentException("String pointer out of bounds.");
			int StringLength = 0;
			int EndOffset = offset;
			while (StringDataSource[EndOffset++] != 0) ++StringLength;
			return Encoding.ASCII.GetString(StringDataSource, (int)offset, StringLength);
		}

		#endregion

		#region Constructor

		/// <summary>Load a compiled QuakeC file from a stream.</summary>
		public QuakeC(ResourceLoader resources, Stream stream) {

			this.Resources = resources;

			BinaryReader QuakeCReader = new BinaryReader(stream);

			#region Load QuakeC Source

			// Check version:
			if (QuakeCReader.ReadInt32() != 6) throw new InvalidDataException("Unsupported version.");

			// Get the CRC:
			this.crc = QuakeCReader.ReadUInt32();

			// Get the directory:
			int[] Directory = new int[12];
			for (int i = 0; i < 12; ++i) Directory[i] = QuakeCReader.ReadInt32();

			// ?
			uint x = QuakeCReader.ReadUInt32();

			// Load statements:
			stream.Seek(Directory[0], SeekOrigin.Begin);
			this.statements = new Statement[Directory[1]];
			for (int i = 0; i < this.statements.Length; ++i) this.statements[i] = new Statement(this, stream);

			// Load functions:
			stream.Seek(Directory[6], SeekOrigin.Begin);
			Function[] Functions = new Function[Directory[7]];
			for (int i = 0; i < Functions.Length; ++i) Functions[i] = new Function(this, stream, i);
			this.functions = new FunctionCollection(Functions);

			// Load string data:
			stream.Seek(Directory[8], SeekOrigin.Begin);
			this.GlobalStringData = QuakeCReader.ReadBytes(Directory[9]);

			// Load definitions data:
			stream.Seek(Directory[10], SeekOrigin.Begin);
			//HACK: +8 to allow for out-of-range data access on the final value if read as a vector.
			this.GlobalData = new VariableValue[Directory[11] + 8];
			for (int i = 0; i < Directory[11]; ++i) GlobalData[i] = new VariableValue(QuakeCReader.ReadInt32());
			
			// Load globals:
			stream.Seek(Directory[2], SeekOrigin.Begin);
			Definition[] Globals = new Definition[Directory[3]];
			for (int i = 0; i < Globals.Length; ++i) Globals[i] = new Definition(this, stream, true);
			this.globals = new DefinitionCollection(Globals);

			// Load fields:
			stream.Seek(Directory[4], SeekOrigin.Begin);
			Definition[] Fields = new Definition[Directory[5]];
			for (int i = 0; i < Fields.Length; ++i) Fields[i] = new Definition(this, stream, false);
			this.fields = new DefinitionCollection(Fields);


			// Misc. setup stuff:
			this.EdictStringData = new byte[] { 0 };

			this.edictSize = 0;
			foreach (Definition Def in this.fields) this.edictSize += Def.Type == Definition.DataType.Vector ? 3 : 1;

			#endregion

		}

		#endregion
	}
}
