using System;
using System.IO;
using System.Text;
using System.Runtime.InteropServices;
using System.ComponentModel;
using System.Collections;
using System.Collections.Generic;

namespace id.Quake.Files {
	public partial class QuakeC {

		/// <summary>Defines a collection of variable definitions.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class DefinitionCollection : IEnumerable {
			
			private readonly Definition[] definitions;
			private Dictionary<string, Definition> DefinitionNameTable;

			/// <summary>Creates an instance of the collection from an array of definitions.</summary>
			public DefinitionCollection(Definition[] definitions) {
				this.definitions = definitions;
				this.DefinitionNameTable = new Dictionary<string, Definition>(this.definitions.Length);
				List<string> AlreadyAdded = new List<string>();
				foreach (Definition Def in this.definitions) {
					if (AlreadyAdded.Contains(Def.Name)) {
						if (this.DefinitionNameTable.ContainsKey(Def.Name)) this.DefinitionNameTable.Remove(Def.Name);
					} else {
						this.DefinitionNameTable.Add(Def.Name, Def);
						AlreadyAdded.Add(Def.Name);	
					}					
				}

			}
			/// <summary>Gets an enumerator for the collection.</summary>
			public IEnumerator GetEnumerator() {
				return this.definitions.GetEnumerator();
			}

			/// <summary>Gets the number of definitions in the collection.</summary>
			public int Count {
				get { return this.definitions.Length; }
			}

			/// <summary>Gets a definition by its name.</summary>
			/// <param name="name">The name of the definition to search for (case sensitive).</param>
			public Definition this[string name] {
				get {
					if (!this.DefinitionNameTable.ContainsKey(name)) throw new InvalidOperationException("Definition '" + name + "' not found.");
					return this.DefinitionNameTable[name];

				}
			}

			/// <summary>Gets a definition by its index.</summary>
			/// <param name="index">The index of the definition to retrieve.</param>
			public Definition this[int index] {
				get { return this.definitions[index]; }
			}
		}

		/// <summary>Defines a variable definition.</summary>
		[TypeConverter(typeof(ExpandableObjectConverter))]
		public class Definition {

			private readonly QuakeC Source;

			/// <summary>Defines the underlying type of the definition.</summary>
			public enum DataType : ushort {
				/// <summary>Empty definition.</summary>
				Void,
				/// <summary>A string of characters.</summary>
				String,
				/// <summary>A single-precision floating point number.</summary>
				Float,
				/// <summary>A vector made of three single-precision floating components.</summary>
				Vector,
				/// <summary>An entity (referenced by index).</summary>
				Entity,
				/// <summary>A field definition for accessing entity members.</summary>
				Field,
				/// <summary>A function.</summary>
				Function,
				/// <summary>A pointer of some description.</summary>
				Pointer
			}

			#region Properties

			private readonly DataType type;
			/// <summary>Gets the data type.</summary>
			public DataType Type {
				get { return (DataType)((ushort)this.type & 0x7FFF); }
			}

			/// <summary>Returns true if the definition needs to be saved in savegame files.</summary>
			public bool RequiresSaving {
				get { return ((ushort)this.type & 0x8000) != 0; }
			}

			private readonly bool isGlobal;
			/// <summary>Returns true if the definition is a global, false if it's a field.</summary>
			public bool IsGlobal {
				get { return this.isGlobal; }
			}

			private readonly ushort offset;
			/// <summary>Gets the offset to this definition's data.</summary>
			public ushort Offset {
				get { return this.offset; }
			}

			private readonly int name;
			/// <summary>Gets the name of the definition.</summary>
			public string Name { get { return this.Source.GetString(this.name); } }

			/// <summary>Gets a formatted string representing the definition.</summary>
			public override string ToString() {
				return this.Type.ToString() + " " + this.Name;
			}

			#endregion			

			#region Constructor

			/// <summary>Read a definition from a compiled QuakeC file.</summary>
			/// <param name="source">The source QuakeC file.</param>
			/// <param name="stream">The stream to read from.</param>
			/// <param name="isGlobal">True if this is a global definition, false if it's a field definition.</param>
			public Definition(QuakeC source, Stream stream, bool isGlobal) {
				this.isGlobal = isGlobal;
				this.Source = source;
				BinaryReader DefinitionReader = new BinaryReader(stream);
				this.type = (DataType)DefinitionReader.ReadUInt16();
				this.offset = DefinitionReader.ReadUInt16();
				this.name = DefinitionReader.ReadInt32();
			}

			#endregion

			/// <summary>Get or set a value from this definition.</summary>
			/// <remarks>Only use this method to retrieve values of global variables.</remarks>
			public object Value {
				get {
					if (!this.isGlobal) throw new InvalidOperationException("You cannot use this method to return field values.");
					switch (this.Type) {
						case DataType.Float:
							return this.Source.GlobalData[this.offset].Float;
						case DataType.Function:
							return this.Source.functions[this.Source.GlobalData[this.offset].Integer];
						case DataType.String:
							return this.Source.GetString(this.Source.GlobalData[this.offset].Integer);
						case DataType.Vector:
							return new Vector3(
								this.Source.GlobalData[this.offset + 0].Float,
								this.Source.GlobalData[this.offset + 1].Float,
								this.Source.GlobalData[this.offset + 2].Float);
						case DataType.Void:
							return null;
						case DataType.Entity:
							return this.Source.GlobalData[this.offset].Integer;
						default:
							throw new InvalidOperationException();
					}
				}
				set {
					if (!this.isGlobal) throw new InvalidOperationException("You cannot use this method to set field values.");
					switch (this.Type) {
						case DataType.Float:
							this.Source.GlobalData[this.offset].Float = Convert.ToSingle(value);
							break;
						case DataType.Function:
							this.Source.GlobalData[this.offset].Integer = ((Function)value).Index;
							break;
						case DataType.Vector:
							Vector3 V = (Vector3)value;
							this.Source.GlobalData[this.offset + 0].Float = V.X;
							this.Source.GlobalData[this.offset + 1].Float = V.Y;
							this.Source.GlobalData[this.offset + 2].Float = V.Z;
							break;
						case DataType.Entity:
							this.Source.GlobalData[this.offset].Integer = ((Edict)value).Index;
							break;
						default:
							throw new InvalidOperationException();
					}
				}
			}
		}
	}
}
