using System;
using System.IO;
using System.Text;
using System.Collections.Generic;

namespace id.Quake.Packages {

	/// <summary>Defines a pack file as used by Quake 1 and Quake 2.</summary>
	public class Pack : IPackage, IDisposable {


		#region Types

		/// <summary>Represents an entry in the pack file.</summary>
		struct Entry {

			private readonly Pack pack;
			/// <summary>Gets the pack file that the entry is contained within.</summary>
			public Pack Pack {
				get { return this.pack; }
			}

			private readonly string name;
			/// <summary>Gets the name of the entry, including extension.</summary>
			public string Name {
				get { return this.name; }
			}

			private uint offset;
			/// <summary>Gets the offset of the entry within the pack file in bytes.</summary>
			public uint Offset {
				get { return this.offset; }
			}

			private uint size;
			/// <summary>Gets the size of the entry in bytes.</summary>
			public uint Size {
				get { return this.size; }
			}

			/// <summary>Creates an instance of the entry class.</summary>
			/// <param name="pack">The pack file containing the entry's data.</param>
			public Entry(Pack pack) {
				this.pack = pack;
				BinaryReader EntryReader = new BinaryReader(pack.BaseStream);
				this.name = Encoding.ASCII.GetString(EntryReader.ReadBytes(56)).Split('\0')[0];
				this.offset = EntryReader.ReadUInt32();
				this.size = EntryReader.ReadUInt32();
			}

			/// <summary>Read the data of the pack file entry as an array of bytes.</summary>
			/// <returns>The entry as an array of bytes.</returns>
			public byte[] ReadBytes() {
				this.Pack.BaseStream.Seek(this.Offset, SeekOrigin.Begin);
				byte[] Output = new byte[this.Size];
				if (this.Pack.BaseStream.Read(Output, 0, Output.Length) != Output.Length) throw new IOException("Could not read the entire pack file entry.");
				return Output;
			}

			/// <summary>Read the data of the pack file entry as a stream.</summary>
			/// <returns>The entry as a stream.</returns>
			public Stream ReadStream() {
				return new MemoryStream(this.ReadBytes());
			}

			public override string ToString() {
				return this.Name;
			}
		}


		#endregion

		#region Properties

		/// <summary>Offset of the pack file directory.</summary>
		private readonly uint DirectoryOffset;

		/// <summary>Size of the pack file directory.</summary>
		private readonly uint DirectorySize;

		/// <summary>Table used to quickly look up directory entries.</summary>
		private readonly Dictionary<string, Entry> DirectoryLookup;

		private readonly Stream BaseStream;

		#endregion

		#region Constructor

		public Pack(Stream s) {
			this.BaseStream = s;
			BinaryReader PackLoader = new BinaryReader(this.BaseStream);
			if (Encoding.ASCII.GetString(PackLoader.ReadBytes(4)) != "PACK") throw new InvalidDataException("Invalid pack file header.");
			this.DirectoryOffset = PackLoader.ReadUInt32();
			this.DirectorySize = PackLoader.ReadUInt32();
			this.BaseStream.Seek(this.DirectoryOffset, SeekOrigin.Begin);
			this.DirectoryLookup = new Dictionary<string, Entry>();
			while (this.BaseStream.Position != this.DirectoryOffset + this.DirectorySize) {
				Entry E = new Entry(this);
				string CaseInsensitiveName = E.Name.ToLowerInvariant().Replace('\\', '/');
				if (this.DirectoryLookup.ContainsKey(CaseInsensitiveName)) this.DirectoryLookup.Remove(CaseInsensitiveName);
				this.DirectoryLookup.Add(CaseInsensitiveName, E);
			}
		}

		#endregion

		#region IPackage Members

		public Stream OpenFile(string filename) {
			Entry E;
			if (this.DirectoryLookup.TryGetValue(filename.ToLowerInvariant().Replace('\\', '/'), out E)) return E.ReadStream();
			throw new FileNotFoundException();
		}

		public string[] GetFiles() {
			string[] Names = new string[this.DirectoryLookup.Count];
			int i = 0;
			foreach (Entry E in this.DirectoryLookup.Values) Names[i++] = E.Name;
			return Names;
		}

		public bool FileExists(string filename) {
			return this.DirectoryLookup.ContainsKey(filename.ToLowerInvariant().Replace('\\', '/'));
		}


		#endregion

		#region IDisposable Members

		public void Dispose() {
			if (this.BaseStream != null) {
				this.BaseStream.Close();
				this.BaseStream.Dispose();
			}
		}

		#endregion

	}
}
