using System;
using QuakeC = id.Quake.Files.QuakeC;
using System.Text;

namespace id.Quake {

	/// <summary>Defines an edict.</summary>
	public class Edict {

		#region Types

		/// <summary>Describes the state of the edict.</summary>
		[Flags()]
		public enum EdictFlags {
			Fly = 1,
			Swim = 2,
			Conveyor = 4,
			Client = 8,
			InWater = 16,
			Monster = 32,
			GodMode = 64,
			NoTarget = 128,
			Item = 256,
			OnGround = 512,
			PartialGround = 1024,
			WaterJump = 2048,
			JumpReleased = 4096,
		}

		#endregion

		#region Private Members

		/// <summary>Reference to all of the other edicts.</summary>
		private readonly EdictCollection Edicts;

		#endregion

		#region Properties

		private readonly int index;
		/// <summary>Gets the index of this edict.</summary>
		public int Index {
			get { return this.index; }
		}

		private readonly PropertyCollection properties;
		/// <summary>Gets the collection of properties for this entity.</summary>
		public PropertyCollection Properties {
			get { return this.properties; }
		}

		public override string ToString() {
			return "Edict[" + this.index + "] = " + this.properties.ToString();
		}

		#endregion

		#region Constructor

		public Edict(EdictCollection edicts, int index) {
			this.Edicts = edicts;
			this.index = index;
			this.properties = new PropertyCollection(this);
		}

		#endregion

		#region PropertyCollection

		public class PropertyCollection {

			/// <summary>Source edict for the property collection.</summary>
			private readonly Edict SourceEdict;

			public PropertyCollection(Edict source) {
				this.SourceEdict = source;
			}

			#region Indexers

			/// <summary>Gets or sets a field by its name.</summary>
			/// <param name="field">Name of the field to access.</param>
			public object this[string field] {
				get { return this[this.SourceEdict.Edicts.QuakeCSource.Fields[field]]; }
				set { this[this.SourceEdict.Edicts.QuakeCSource.Fields[field]] = value; }
			}

			/// <summary>Gets or sets a field by its definition.</summary>
			/// <param name="field">Field to access.</param>
			public object this[Files.QuakeC.Definition field] {
				get {
					if (field.IsGlobal) throw new ArgumentException("Cannot access global variables as edict fields.");
					int Offset = (this.SourceEdict.index * this.SourceEdict.Edicts.QuakeCSource.EdictSize) + field.Offset;
					switch (field.Type) {
						case QuakeC.Definition.DataType.Float:
							return this.SourceEdict.Edicts.EdictData[Offset].Float;
						case QuakeC.Definition.DataType.Function:
							return this.SourceEdict.Edicts.QuakeCSource.Functions[this.SourceEdict.Edicts.EdictData[Offset].Integer];
						case QuakeC.Definition.DataType.String:
							return this.SourceEdict.Edicts.QuakeCSource.GetString(this.SourceEdict.Edicts.EdictData[Offset].Integer);
						case QuakeC.Definition.DataType.Vector:
							return new Vector3(
								this.SourceEdict.Edicts.EdictData[Offset + 0].Float,
								this.SourceEdict.Edicts.EdictData[Offset + 1].Float,
								this.SourceEdict.Edicts.EdictData[Offset + 2].Float);
						case QuakeC.Definition.DataType.Void:
							return null;
						default:
							throw new InvalidOperationException();
					}
				}
				set {
					if (field.IsGlobal) throw new ArgumentException("Cannot access global variables as edict fields.");
					int Offset = (this.SourceEdict.index * this.SourceEdict.Edicts.QuakeCSource.EdictSize) + field.Offset;
					switch (field.Type) {
						case QuakeC.Definition.DataType.Float:
							this.SourceEdict.Edicts.EdictData[Offset].Float = Convert.ToSingle(value);
							break;
						case QuakeC.Definition.DataType.Function:
							this.SourceEdict.Edicts.EdictData[Offset].Integer = ((QuakeC.Function)value).Index;
							break;
						case QuakeC.Definition.DataType.Vector:
							Vector3 V = (Vector3)value;
							this.SourceEdict.Edicts.EdictData[Offset + 0].Float = V.X;
							this.SourceEdict.Edicts.EdictData[Offset + 1].Float = V.Y;
							this.SourceEdict.Edicts.EdictData[Offset + 2].Float = V.Z;
							break;
						case QuakeC.Definition.DataType.String:
							this.SourceEdict.Edicts.EdictData[Offset].Integer = (int)value;
							break;
						default:
							throw new InvalidOperationException();
					}
				}
			}

			#endregion

			public override string ToString() {
				StringBuilder PropertiesString = new StringBuilder();
				PropertiesString.AppendLine("{");
				foreach (QuakeC.Definition Def in this.SourceEdict.Edicts.QuakeCSource.Fields) {
					int Offset = this.SourceEdict.index * this.SourceEdict.Edicts.QuakeCSource.EdictSize + Def.Offset;
					if (this.SourceEdict.Edicts.EdictData[Offset].Integer != 0) {
						PropertiesString.AppendLine(string.Format("\"{0}\" = \"{1}\"", Def.Name, this[Def]));
					}
				}
				PropertiesString.Append("}");
				return PropertiesString.ToString();
			}

		}

		#endregion

	}
}
