using System;
using QuakeC = id.Quake.Files.QuakeC;

namespace id.Quake {
	public partial class VirtualMachine {

		#region Types

		/// <summary>Working variables are stored in the globals and used to pass or return values.</summary>
		/// <remarks>Parameters go up in steps of three to allow for three vector components.</remarks>
		private enum WorkingVariableOffsets: ushort {
			Null = 0,
			Return = 1,
			Parameter0 = 4, Parameter1 = 7, Parameter2 = 10, Parameter3 = 13, Parameter4 = 16, Parameter5 = 19, Parameter6 = 22, Parameter7 = 25,
			Reserved = 28,
		}

		/// <summary>Used for quick access to reading/writing working variable data.</summary>
		private class WorkingVariable {

			/// <summary>Index of data within global block.</summary>
			private readonly ushort Index;

			/// <summary>Owner virtual machine.</summary>
			private readonly VirtualMachine VirtualMachine;

			public WorkingVariable(VirtualMachine vm, ushort index) {
				this.VirtualMachine = vm;
				this.Index = index;
			}

			#region Base Properties

			public int Integer {
				get { return this.VirtualMachine.quakeCSource.GlobalData[this.Index].Integer; }
				set { this.VirtualMachine.quakeCSource.GlobalData[this.Index].Integer = value; }
			}

			public float Float {
				get { return this.VirtualMachine.quakeCSource.GlobalData[this.Index].Float; }
				set { this.VirtualMachine.quakeCSource.GlobalData[this.Index].Float = value; }
			}

			public Vector3 Vector {
				get {
					return new Vector3(
						this.VirtualMachine.quakeCSource.GlobalData[this.Index + 0].Float,
						this.VirtualMachine.quakeCSource.GlobalData[this.Index + 1].Float,
						this.VirtualMachine.quakeCSource.GlobalData[this.Index + 2].Float
					);
				}
				set {
					this.VirtualMachine.quakeCSource.GlobalData[this.Index + 0].Float = value.X;
					this.VirtualMachine.quakeCSource.GlobalData[this.Index + 1].Float = value.Y;
					this.VirtualMachine.quakeCSource.GlobalData[this.Index + 2].Float = value.Z;
				}
			}

			#endregion

			#region Extended Properties

			public bool Boolean {
				get { return this.Float != 0f; }
				set { this.Float = value ? 1f : 0f; }
			}

			public string String {
				get { return this.VirtualMachine.quakeCSource.GetString(this.Integer); }
			}

			public Edict Edict {
				get { return this.VirtualMachine.edicts[this.Integer]; }
				set { this.Integer = value.Index; }
			}

			public QuakeC.Function Function {
				get { return this.VirtualMachine.quakeCSource.Functions[this.Integer]; }
			}

			#endregion
		}

		#endregion


		#region Working Variables

		/// <summary>Gets or sets return data.</summary>
		private readonly WorkingVariable Return;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter0;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter1;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter2;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter3;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter4;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter5;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter6;
		/// <summary>Gets or sets parameter data.</summary>
		private readonly WorkingVariable Parameter7;

		#endregion

		#region Important definitions.
		
		/// <summary>The 'self' definition.</summary>
		private readonly QuakeC.Definition Self;
		/// <summary>The 'other' definition.</summary>
		private readonly QuakeC.Definition Other;
		/// <summary>The 'time' definition.</summary>
		private readonly QuakeC.Definition Time;

		#endregion


		private Edict SelfEdict {
			get { return this.edicts[(int)this.Self.Value]; }
		}

	}
}
